import { ClientBase, Pool, PoolClient } from 'pg';
import { TransactionalLogger } from './logger';
/**
 * Sleep for a given amount of milliseconds
 * @param milliseconds The time in milliseconds to sleep
 * @returns The (void) promise to await
 */
export declare const sleep: (milliseconds: number) => Promise<void>;
/**
 * Run a promise but make sure to wait only a maximum amount of time for it to finish.
 * @param promise The promise to execute
 * @param timeoutMs The amount of time in milliseconds to wait for the promise to finish
 * @param failureMessage The message for the error if the timeout was reached
 * @returns The promise return value or a timeout error is thrown
 */
export declare const awaitWithTimeout: <T>(promise: () => Promise<T>, timeoutMs: number, failureMessage?: string) => Promise<T>;
/**
 * PostgreSQL available isolation levels. The isolation level "Read uncommitted"
 * is the same as "Read committed" in PostgreSQL. And the readonly variants are
 * not usable as the inbox message must be marked as processed or the
 * "finished_attempts" counter is updated for the message.
 */
export declare enum IsolationLevel {
    Serializable = "SERIALIZABLE",
    RepeatableRead = "REPEATABLE READ",
    ReadCommitted = "READ COMMITTED"
}
/**
 * Open a transaction and execute the callback as part of the transaction.
 * @param client The PostgreSQL database client
 * @param callback The callback to execute DB commands with.
 * @param isolationLevel The database transaction isolation level. Falls back to the default PostgreSQL transaction level if not provided.
 * @returns The result of the callback (if any).
 * @throws Any error from the database or the callback.
 */
export declare const executeTransaction: <T>(client: ClientBase | PoolClient, callback: (client: ClientBase) => Promise<T>, isolationLevel?: IsolationLevel) => Promise<T>;
/**
 * Creates a PoolClient from the given pool.
 * @param pool A PostgreSQL database pool from which clients can be created.
 * @param logger A logger that will be registered for the on-error event of the client.
 * @returns The PoolClient object
 */
export declare const getClient: (pool: Pool, logger?: TransactionalLogger) => Promise<PoolClient>;
//# sourceMappingURL=utils.d.ts.map