"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createMultiConcurrencyController = void 0;
const create_discriminating_mutex_concurrency_controller_1 = require("./create-discriminating-mutex-concurrency-controller");
const create_full_concurrency_controller_1 = require("./create-full-concurrency-controller");
const create_mutex_concurrency_controller_1 = require("./create-mutex-concurrency-controller");
/**
 * Use different mutex controllers depending on the desired concurrency level
 * for different messages. If a discriminating mutex should be used, the
 * discriminator function must be supplied as well.
 * @param getConcurrencyType Implements the logic which concurrency controller should be used e.g. based on the aggregate and message type
 * @param discriminator The discriminator to find or create a mutex for when using the discriminating mutex.
 * @returns The controller to acquire and release the mutex for a specific discriminator
 */
const createMultiConcurrencyController = (getConcurrencyType, discriminator) => {
    const fullConcurrencyController = (0, create_full_concurrency_controller_1.createFullConcurrencyController)();
    const mutexController = (0, create_mutex_concurrency_controller_1.createMutexConcurrencyController)();
    const discriminatingMutexController = discriminator
        ? (0, create_discriminating_mutex_concurrency_controller_1.createDiscriminatingMutexConcurrencyController)(discriminator)
        : undefined;
    return {
        /** Acquire a lock (if any) and return a function to release it. */
        acquire: (message) => {
            switch (getConcurrencyType(message)) {
                case 'full-concurrency':
                    return fullConcurrencyController.acquire(message);
                case 'mutex':
                    return mutexController.acquire(message);
                case 'discriminating-mutex':
                    if (!discriminatingMutexController) {
                        throw new Error('A discriminating mutex controller was not configured.');
                    }
                    return discriminatingMutexController.acquire(message);
            }
        },
        /** Cancel all controllers. */
        cancel: () => {
            fullConcurrencyController.cancel();
            mutexController.cancel();
            discriminatingMutexController === null || discriminatingMutexController === void 0 ? void 0 : discriminatingMutexController.cancel();
        },
    };
};
exports.createMultiConcurrencyController = createMultiConcurrencyController;
//# sourceMappingURL=create-multi-concurrency-controller.js.map