import { ClientBase } from 'pg';
import { TransactionalLogger } from '../common/logger';
import { InboxMessage, OutboxMessage } from '../common/message';
import { InboxConfig } from './inbox-listener';
/**
 * Initialize the inbox message storage to store incoming messages in the inbox table.
 * @param config The configuration object that defines the values on how to connect to the database and general settings.
 * @param logger A logger instance for logging trace up to error logs
 * @returns The function to store the inbox message data to the database and the shutdown action.
 */
export declare const initializeInboxMessageStorage: (config: Pick<InboxConfig, 'pgConfig' | 'settings'>, logger: TransactionalLogger) => [storeInboxMessage: (message: OutboxMessage) => Promise<void>, shutdown: () => Promise<void>];
/**
 * This function increases the "started_attempts" for the inbox message by one
 * in the inbox table. This number can then be compared to the
 * "finished_attempts" number which is only increased when a message processing
 * exception was correctly handled or an error was caught. A difference between
 * the two can only happen if the service crashes after increasing the
 * "started_attempts" but before successfully marking the message as done
 * (success case) or catching an error (error case). If the "started_attempts"
 * and the "finished_attempts" field differ by more than one, the chances are
 * high that this message is causing a service crash.
 * @param message The inbox message for which to check/increase the "started_attempts".
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema and poisonous settings.
 * @returns The "started_attempts" (one or higher) and the "finished_attempts" (zero or higher). Or undefined if the message was not found.
 */
export declare const poisonousMessageUpdate: ({ id }: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<PoisonousCheck | undefined>;
export interface PoisonousCheck {
    startedAttempts: number;
    finishedAttempts: number;
}
/**
 * Make sure the inbox item was not and is not currently being worked on. And
 * set the started_attempts, finished_attempts and processed_at values for the
 * WAL message. As the inbox listener does not run in the same transaction as
 * the message handler code there is a small chance that the handler code
 * succeeds but the WAL inbox message was not acknowledged. This takes care of
 * such cases by ensuring that the message was not yet processed.
 * @param message The inbox message to check and later process.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 * @returns 'INBOX_MESSAGE_NOT_FOUND' if the message was not found, 'ALREADY_PROCESSED' if it was processed, and otherwise assigns the properties to the message and returns true.
 */
export declare const verifyInbox: (message: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<true | 'INBOX_MESSAGE_NOT_FOUND' | 'ALREADY_PROCESSED'>;
/**
 * Acknowledge that the inbox message was handled.
 * @param message The inbox message to acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
export declare const ackInbox: ({ id }: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<void>;
/**
 * Not acknowledging the inbox message means increasing the finished_attempts counter by one
 * @param message The inbox message to NOT acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
export declare const nackInbox: ({ id }: InboxMessage, client: ClientBase, { settings }: Pick<InboxConfig, 'settings'>) => Promise<void>;
//# sourceMappingURL=inbox-message-storage.d.ts.map