"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeGeneralOutboxMessageStorage = exports.initializeOutboxMessageStorage = void 0;
const uuid_1 = require("uuid");
const error_1 = require("../common/error");
/**
 * Pre-configure the specific kind of outbox message to generate and receive a
 * function that can be called to actually store the specific outbox message.
 * @param aggregateType The aggregate root type that was responsible to create this message.
 * @param messageType The type of message that happened on the aggregate type.
 * @param config The configuration object that defines the outbox table schema.
 * @returns The function to store the outbox message data to the database.
 */
const initializeOutboxMessageStorage = (aggregateType, messageType, { settings: { dbSchema, dbTable } }) => {
    /**
     * Function to store the outbox message data to the database.
     * @param aggregateId The identifier of the aggregate.
     * @param payload The actual message payload that should be sent.
     * @param dbClient The database client that should have an active transaction to insert the outbox data along with the business logic transaction.
     * @param metadata Optional metadata that is/was used for the actual message transfer
     * @returns The outbox message data that was stored in the database.
     * @throws An error if the outbox message could not be created.
     */
    return (aggregateId, payload, dbClient, metadata) => __awaiter(void 0, void 0, void 0, function* () {
        return storeOutboxMessage(dbSchema, dbTable, aggregateId, aggregateType, messageType, payload, dbClient, metadata);
    });
};
exports.initializeOutboxMessageStorage = initializeOutboxMessageStorage;
/**
 * Create a general outbox message storage with only the DB settings to generate
 * and receive a function that can be called to actually store the specific
 * outbox message while supplying also the aggregate type and message type.
 * @param config The configuration object that defines the outbox table schema.
 * @returns The function to store the outbox message data to the database.
 */
const initializeGeneralOutboxMessageStorage = ({ settings: { dbSchema, dbTable }, }) => {
    /**
     * Function to store the outbox message data to the database.
     * @param aggregateId The identifier of the aggregate.
     * @param aggregateType The aggregate root type that was responsible to create this message.
     * @param messageType The type of message that happened on the aggregate type.
     * @param payload The actual message payload that should be sent.
     * @param dbClient The database client that should have an active transaction to insert the outbox data along with the business logic transaction.
     * @param metadata Optional metadata that is/was used for the actual message transfer
     * @returns The outbox message data that was stored in the database.
     * @throws An error if the outbox message could not be created.
     */
    return (aggregateId, aggregateType, messageType, payload, dbClient, metadata) => __awaiter(void 0, void 0, void 0, function* () {
        return storeOutboxMessage(dbSchema, dbTable, aggregateId, aggregateType, messageType, payload, dbClient, metadata);
    });
};
exports.initializeGeneralOutboxMessageStorage = initializeGeneralOutboxMessageStorage;
const storeOutboxMessage = (dbSchema, dbTable, aggregateId, aggregateType, messageType, payload, dbClient, metadata) => __awaiter(void 0, void 0, void 0, function* () {
    const outboxId = (0, uuid_1.v4)();
    const outboxResult = yield dbClient.query(
    /* sql*/ `
    INSERT INTO ${dbSchema}.${dbTable}
      (id, aggregate_type, aggregate_id, message_type, payload, metadata)
      VALUES ($1, $2, $3, $4, $5, $6)
    RETURNING id, created_at;`, [outboxId, aggregateType, aggregateId, messageType, payload, metadata]);
    const attemptedMessage = {
        aggregateType,
        aggregateId,
        messageType,
        payload,
        metadata,
        id: 'unknown',
        createdAt: 'unknown',
    };
    if (!outboxResult.rowCount || outboxResult.rowCount < 1) {
        throw new error_1.MessageError('Could not insert the message into the outbox!', attemptedMessage);
    }
    // Immediately delete the outbox entry - it was already written to the WAL
    yield dbClient.query(
    /* sql*/ `
    DELETE FROM ${dbSchema}.${dbTable} WHERE id = $1;`, [outboxId]);
    return Object.assign(Object.assign({}, attemptedMessage), { id: outboxId, createdAt: outboxResult.rows[0].created_at });
});
//# sourceMappingURL=outbox-message-storage.js.map