import { Pgoutput } from 'pg-logical-replication';
import { TransactionalLogger } from '../common/logger';
import { OutboxMessage } from '../common/message';
import { ConcurrencyController } from '../concurrency-controller/concurrency-controller';
import { ListenerRestartTimeStrategy } from '../strategies/listener-restart-time-strategy';
import { MessageProcessingTimeoutStrategy } from '../strategies/message-processing-timeout-strategy';
import { TransactionalOutboxInboxConfig } from './config';
/** Optional strategies to provide custom logic for handling specific scenarios */
export interface TransactionalStrategies {
    /**
     * Define the concurrency strategy - defaults to using a mutex to guarantee
     * sequential message processing.
     */
    concurrencyStrategy: ConcurrencyController;
    /**
     * Defines the message processing timeout strategy. By default, it uses the
     * configured messageProcessingTimeout or falls back to a 15-second timeout.
     */
    messageProcessingTimeoutStrategy: MessageProcessingTimeoutStrategy;
    /**
     * Returns the time in milliseconds after an error was caught until the
     * listener should try to restart itself.
     */
    listenerRestartTimeStrategy: ListenerRestartTimeStrategy;
}
/**
 * Initiate the outbox or inbox listener to listen for WAL messages.
 * @param config The replication connection settings and general replication settings
 * @param messageHandler The message handler that handles the outbox/inbox message
 * @param errorHandler A handler that can decide if the WAL message should be acknowledged (permanent_error) or not (transient_error which restarts the logical replication listener)
 * @param logger A logger instance for logging trace up to error logs
 * @param strategies Strategies to provide custom logic for handling specific scenarios
 * @returns A function to stop the logical replication listener
 */
export declare const createLogicalReplicationListener: <T extends OutboxMessage>({ pgReplicationConfig, settings }: TransactionalOutboxInboxConfig, messageHandler: (message: T) => Promise<void>, errorHandler: (message: T, err: Error) => Promise<boolean>, logger: TransactionalLogger, strategies: TransactionalStrategies, listenerType: 'inbox' | 'outbox') => [shutdown: {
    (): Promise<void>;
}];
/** Get and map the outbox/inbox message if the WAL log entry is such a message. Otherwise returns undefined. */
declare const getRelevantMessage: <T extends OutboxMessage>(log: Pgoutput.Message, { dbSchema, dbTable }: TransactionalOutboxInboxConfig['settings']) => T | undefined;
/** Maps the WAL log entry to an outbox or inbox message */
declare const mapMessage: <T extends OutboxMessage>(input: unknown) => T | undefined;
/**
 * This export is _only_ done for unit tests as the createLogicalReplicationListener
 * function is otherwise very hard to unit test. Exports work only for jest tests!
 */
export declare const __only_for_unit_tests__: {
    getRelevantMessage?: typeof getRelevantMessage;
    mapMessage?: typeof mapMessage;
};
export {};
//# sourceMappingURL=logical-replication-listener.d.ts.map